#!/usr/bin/env python
# coding=utf-8

"""
Check whether the setup of aeneas was successful.

Running this script makes sense only
if you git-cloned the original GitHub repository
and/or if you are interested in contributing to the
development of aeneas.
"""

from __future__ import absolute_import
from __future__ import print_function
import os
import sys

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL 3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

ANSI_ERROR = u"\033[91m"
ANSI_OK = u"\033[92m"
ANSI_WARNING = u"\033[93m"
ANSI_END = u"\033[0m"

def is_posix():
    return os.name == "posix"

def print_error(msg):
    if is_posix():
        print(u"%s[ERRO] %s%s" % (ANSI_ERROR, msg, ANSI_END))
    else:
        print(u"[ERRO] %s" % (msg))

def print_info(msg):
    print(u"[INFO] %s" % (msg))

def print_success(msg):
    if is_posix():
        print(u"%s[INFO] %s%s" % (ANSI_OK, msg, ANSI_END))
    else:
        print(u"[INFO] %s" % (msg))

def print_warning(msg):
    if is_posix():
        print(u"%s[WARN] %s%s" % (ANSI_WARNING, msg, ANSI_END))
    else:
        print(u"[WARN] %s" % (msg))

def check_import():
    try:
        import aeneas
        print_success(u"aeneas OK")
        return False 
    except ImportError:
        print_error(u"aeneas ERROR")
        print_info(u"  Unable to load the aeneas Python package")
        print_info(u"  This error is probably caused by:")
        print_info(u"    A. you did not download/git-clone the aeneas package properly; or")
        print_info(u"    B. you did not install the required Python packages:")
        print_info(u"      1. BeautifulSoup4")
        print_info(u"      2. lxml")
        print_info(u"      3. numpy")
    except Exception as e:
        print_error(e)
    return True

def main():
    # first, check we can import aeneas module
    if check_import():
        sys.exit(1)

    # then, run the built-in diagnostics
    from aeneas.diagnostics import Diagnostics
    errors, warnings, c_ext_warnings = Diagnostics.check_all()
    if errors:
        sys.exit(1)
    print_info(u"")
    if c_ext_warnings:
        print_warning(u"All required dependencies are met but at least one available Python C extension is not compiled")
        print_info(u"You can still run aeneas but it will be slower")
        print_info(u"Enjoy running aeneas!")
        sys.exit(2)
    else:
        print_success(u"All required dependencies are met and all available Python C extensions are compiled")
        print_info(u"Enjoy running aeneas!")
        sys.exit(0)



if __name__ == '__main__':
    main()






#!/usr/bin/env python
# coding=utf-8

"""
A structure representing a task, that is,
an audio file and a list of text fragments
to be synchronized.
"""

from __future__ import absolute_import
from __future__ import print_function
import os

from aeneas.audiofile import AudioFile
from aeneas.configurationobject import ConfigurationObject
from aeneas.logger import Logger
from aeneas.textfile import TextFile
import aeneas.globalconstants as gc
import aeneas.globalfunctions as gf

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class Task(object):
    """
    A structure representing a task, that is,
    an audio file and a list of text fragments
    to be synchronized.

    :param config_string: the task configuration string
    :type  config_string: string
    :param logger: the logger object
    :type  logger: :class:`aeneas.logger.Logger`

    :raises TypeError: if ``config_string`` is not ``None`` and
                       it is not a Unicode string
    """

    TAG = u"Task"

    def __init__(self, config_string=None, logger=None):
        self.logger = logger or Logger()
        self.identifier = gf.uuid_string()
        self.configuration = None
        self.audio_file_path = None # relative to input container root
        self.audio_file_path_absolute = None # concrete path, file will be read from this!
        self.audio_file = None
        self.text_file_path = None # relative to input container root
        self.text_file_path_absolute = None # concrete path, file will be read from this!
        self.text_file = None
        self.sync_map_file_path = None # relative to output container root
        self.sync_map_file_path_absolute = None # concrete path, file will be written to this!
        self.sync_map = None
        if config_string is not None:
            self.configuration = TaskConfiguration(config_string)

    def _log(self, message, severity=Logger.DEBUG):
        """ Log """
        self.logger.log(message, severity, self.TAG)

    def __unicode__(self):
        msg = [
            u"%s: '%s'" % (gc.RPN_TASK_IDENTIFIER, self.identifier),
            u"Configuration:\n%s" % self.configuration.__unicode__(),
            u"Audio file path: %s" % self.audio_file_path,
            u"Audio file path (absolute): %s" % self.audio_file_path_absolute,
            u"Text file path: %s" % self.text_file_path,
            u"Text file path (absolute): %s" % self.text_file_path_absolute,
            u"Sync map file path: %s" % self.sync_map_file_path,
            u"Sync map file path (absolute): %s" % self.sync_map_file_path_absolute
        ]
        return u"\n".join(msg)

    def __str__(self):
        return gf.safe_str(self.__unicode__())

    @property
    def identifier(self):
        """
        The identifier of the task.

        :rtype: string
        """
        return self.__identifier
    @identifier.setter
    def identifier(self, value):
        self.__identifier = value

    @property
    def audio_file_path_absolute(self):
        """
        The absolute path of the audio file.

        :rtype: string (path)
        """
        return self.__audio_file_path_absolute
    @audio_file_path_absolute.setter
    def audio_file_path_absolute(self, audio_file_path_absolute):
        self.__audio_file_path_absolute = audio_file_path_absolute
        self._populate_audio_file()

    @property
    def text_file_path_absolute(self):
        """
        The absolute path of the text file.

        :rtype: string (path)
        """
        return self.__text_file_path_absolute
    @text_file_path_absolute.setter
    def text_file_path_absolute(self, text_file_path_absolute):
        self.__text_file_path_absolute = text_file_path_absolute
        self._populate_text_file()

    @property
    def sync_map_file_path_absolute(self):
        """
        The absolute path of the sync map file.

        :rtype: string (path)
        """
        return self.__sync_map_file_path_absolute
    @sync_map_file_path_absolute.setter
    def sync_map_file_path_absolute(self, sync_map_file_path_absolute):
        self.__sync_map_file_path_absolute = sync_map_file_path_absolute

    def output_sync_map_file(self, container_root_path=None):
        """
        Output the sync map file for this task.

        If ``container_root_path`` is specified,
        the output sync map file will be created
        at the path obtained by joining
        the ``container_root_path`` and the relative path
        of the sync map inside the container.

        Otherwise, the sync map file will be created at the path
        ``sync_map_file_path_absolute``.

        Return the the path of the sync map file created,
        or ``None`` if an error occurred.

        :param container_root_path: the path to the root directory
                                    for the output container
        :type  container_root_path: string (path)
        :rtype: string (path)
        """
        if self.sync_map is None:
            self._log(u"sync_map is None", Logger.CRITICAL)
            raise TypeError("sync_map object has not been set")

        if (container_root_path is not None) and (self.sync_map_file_path is None):
            self._log(u"The (internal) path of the sync map has been set", Logger.CRITICAL)
            raise TypeError("The (internal) path of the sync map has been set")

        self._log([u"container_root_path is %s", container_root_path])
        self._log([u"self.sync_map_file_path is %s", self.sync_map_file_path])
        self._log([u"self.sync_map_file_path_absolute is %s", self.sync_map_file_path_absolute])

        if (container_root_path is not None) and (self.sync_map_file_path is not None):
            path = os.path.join(container_root_path, self.sync_map_file_path)
        elif self.sync_map_file_path_absolute:
            path = self.sync_map_file_path_absolute
        gf.ensure_parent_directory(path)
        self._log([u"Output sync map to %s", path])

        sync_map_format = self.configuration["o_format"]
        audio_ref = self.configuration["o_smil_audio_ref"]
        page_ref = self.configuration["o_smil_page_ref"]

        self._log([u"sync_map_format is %s", sync_map_format])
        self._log([u"page_ref is %s", page_ref])
        self._log([u"audio_ref is %s", audio_ref])

        self._log(u"Calling sync_map.write...")
        # TODO just pass self.configuration?
        parameters = {
            gc.PPN_TASK_OS_FILE_SMIL_PAGE_REF : page_ref,
            gc.PPN_TASK_OS_FILE_SMIL_AUDIO_REF : audio_ref
        }
        self.sync_map.write(sync_map_format, path, parameters)
        self._log(u"Calling sync_map.write... done")
        return path

    def _populate_audio_file(self):
        """
        Create the ``self.audio_file`` object by reading
        the audio file at ``self.audio_file_path_absolute``.
        """
        self._log(u"Populate audio file...")
        if self.audio_file_path_absolute is not None:
            self._log([u"audio_file_path_absolute is '%s'", self.audio_file_path_absolute])
            self.audio_file = AudioFile(
                file_path=self.audio_file_path_absolute,
                logger=self.logger
            )
            self.audio_file.read_properties()
        else:
            self._log(u"audio_file_path_absolute is None")
        self._log(u"Populate audio file... done")

    def _populate_text_file(self):
        """
        Create the ``self.text_file`` object by reading
        the text file at ``self.text_file_path_absolute``.
        """
        self._log(u"Populate text file...")
        if (
                (self.text_file_path_absolute is not None) and
                (self.configuration["language"] is not None)
            ):
            # TODO just pass self.configuration?
            # the following values might be None
            parameters = {
                gc.PPN_TASK_IS_TEXT_FILE_IGNORE_REGEX : self.configuration["i_t_ignore_regex"],
                gc.PPN_TASK_IS_TEXT_FILE_TRANSLITERATE_MAP : self.configuration["i_t_transliterate_map"],
                gc.PPN_TASK_IS_TEXT_UNPARSED_CLASS_REGEX :self.configuration["i_t_unparsed_class_regex"],
                gc.PPN_TASK_IS_TEXT_UNPARSED_ID_REGEX : self.configuration["i_t_unparsed_id_regex"],
                gc.PPN_TASK_IS_TEXT_UNPARSED_ID_SORT : self.configuration["i_t_unparsed_id_sort"],
                gc.PPN_TASK_OS_FILE_ID_REGEX : self.configuration["o_id_regex"]
            }
            self.text_file = TextFile(
                file_path=self.text_file_path_absolute,
                file_format=self.configuration["i_t_format"],
                parameters=parameters,
                logger=self.logger
            )
            self.text_file.set_language(self.configuration["language"])
        else:
            self._log(u"text_file_path_absolute and/or language is None")
        self._log(u"Populate text file... done")



class TaskConfiguration(ConfigurationObject):
    """
    A structure representing a configuration for a task, that is,
    a series of directives for I/O and processing the task.

    Allowed keys:

    * ``PPN_TASK_CUSTOM_ID``                          or ``custom_id``
    * ``PPN_TASK_DESCRIPTION``                        or ``description``
    * ``PPN_TASK_LANGUAGE``                           or ``language``
    * ``PPN_TASK_ADJUST_BOUNDARY_AFTERCURRENT_VALUE`` or ``aba_aftercurrent_value``
    * ``PPN_TASK_ADJUST_BOUNDARY_ALGORITHM``          or ``aba_algorithm``
    * ``PPN_TASK_ADJUST_BOUNDARY_BEFORENEXT_VALUE``   or ``aba_beforenext_value``
    * ``PPN_TASK_ADJUST_BOUNDARY_OFFSET_VALUE``       or ``aba_offset_value``
    * ``PPN_TASK_ADJUST_BOUNDARY_PERCENT_VALUE``      or ``aba_percent_value``
    * ``PPN_TASK_ADJUST_BOUNDARY_RATE_VALUE``         or ``aba_rate_value``
    * ``PPN_TASK_IS_AUDIO_FILE_DETECT_HEAD_MAX``      or ``i_a_head_max``
    * ``PPN_TASK_IS_AUDIO_FILE_DETECT_HEAD_MIN``      or ``i_a_head_min``
    * ``PPN_TASK_IS_AUDIO_FILE_DETECT_TAIL_MAX``      or ``i_a_tail_max``
    * ``PPN_TASK_IS_AUDIO_FILE_DETECT_TAIL_MIN``      or ``i_a_tail_min``
    * ``PPN_TASK_IS_AUDIO_FILE_HEAD_LENGTH``          or ``i_a_head``
    * ``PPN_TASK_IS_AUDIO_FILE_PROCESS_LENGTH``       or ``i_a_process``
    * ``PPN_TASK_IS_AUDIO_FILE_TAIL_LENGTH``          or ``i_a_tail``
    * ``PPN_TASK_IS_TEXT_FILE_FORMAT``                or ``i_t_format``
    * ``PPN_TASK_IS_TEXT_FILE_IGNORE_REGEX``          or ``i_t_ignore_regex``
    * ``PPN_TASK_IS_TEXT_FILE_TRANSLITERATE_MAP``     or ``i_t_transliterate_map``
    * ``PPN_TASK_IS_TEXT_UNPARSED_CLASS_REGEX``       or ``i_t_unparsed_class_regex``
    * ``PPN_TASK_IS_TEXT_UNPARSED_ID_REGEX``          or ``i_t_unparsed_id_regex``
    * ``PPN_TASK_IS_TEXT_UNPARSED_ID_SORT``           or ``i_t_unparsed_id_sort``
    * ``PPN_TASK_OS_FILE_FORMAT``                     or ``o_format``
    * ``PPN_TASK_OS_FILE_HEAD_TAIL_FORMAT``           or ``o_h_t_format``
    * ``PPN_TASK_OS_FILE_ID_REGEX``                   or ``o_id_regex``
    * ``PPN_TASK_OS_FILE_NAME``                       or ``o_name``
    * ``PPN_TASK_OS_FILE_SMIL_AUDIO_REF``             or ``o_smil_audio_ref``
    * ``PPN_TASK_OS_FILE_SMIL_PAGE_REF``              or ``o_smil_page_ref``

    :param config_string: the job configuration string
    :type  config_string: Unicode string

    :raises TypeError: if ``config_string`` is not ``None`` and
                       it is not a Unicode string
    :raises KeyError: if trying to access a key not listed above
    """

    TAG = u"TaskConfiguration"

    FIELDS = [
        (gc.PPN_TASK_CUSTOM_ID, (None, None, ["custom_id"])),
        (gc.PPN_TASK_DESCRIPTION, (None, None, ["description"])),
        (gc.PPN_TASK_LANGUAGE, (None, None, ["language"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_AFTERCURRENT_VALUE, (None, float, ["aba_aftercurrent_value"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM, (None, None, ["aba_algorithm"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_BEFORENEXT_VALUE, (None, float, ["aba_beforenext_value"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_OFFSET_VALUE, (None, float, ["aba_offset_value"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_PERCENT_VALUE, (None, int, ["aba_percent_value"])),
        (gc.PPN_TASK_ADJUST_BOUNDARY_RATE_VALUE, (None, float, ["aba_rate_value"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_DETECT_HEAD_MAX, (None, float, ["i_a_head_max"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_DETECT_HEAD_MIN, (None, float, ["i_a_head_min"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_DETECT_TAIL_MAX, (None, float, ["i_a_tail_max"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_DETECT_TAIL_MIN, (None, float, ["i_a_tail_min"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_HEAD_LENGTH, (None, float, ["i_a_head"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_PROCESS_LENGTH, (None, float, ["i_a_process"])),
        (gc.PPN_TASK_IS_AUDIO_FILE_TAIL_LENGTH, (None, float, ["i_a_tail"])),
        (gc.PPN_TASK_IS_TEXT_FILE_FORMAT, (None, None, ["i_t_format"])),
        (gc.PPN_TASK_IS_TEXT_FILE_IGNORE_REGEX, (None, None, ["i_t_ignore_regex"])),
        (gc.PPN_TASK_IS_TEXT_FILE_TRANSLITERATE_MAP, (None, None, ["i_t_transliterate_map"])),
        (gc.PPN_TASK_IS_TEXT_UNPARSED_CLASS_REGEX, (None, None, ["i_t_unparsed_class_regex"])),
        (gc.PPN_TASK_IS_TEXT_UNPARSED_ID_REGEX, (None, None, ["i_t_unparsed_id_regex"])),
        (gc.PPN_TASK_IS_TEXT_UNPARSED_ID_SORT, (None, None, ["i_t_unparsed_id_sort"])),
        (gc.PPN_TASK_OS_FILE_FORMAT, (None, None, ["o_format"])),
        (gc.PPN_TASK_OS_FILE_HEAD_TAIL_FORMAT, (None, None, ["o_h_t_format"])),
        (gc.PPN_TASK_OS_FILE_ID_REGEX, (None, None, ["o_id_regex"])),
        (gc.PPN_TASK_OS_FILE_NAME, (None, None, ["o_name"])),
        (gc.PPN_TASK_OS_FILE_SMIL_AUDIO_REF, (None, None, ["o_smil_audio_ref"])),
        (gc.PPN_TASK_OS_FILE_SMIL_PAGE_REF, (None, None, ["o_smil_page_ref"])),
    ]

    def __init__(self, config_string=None):
        super(TaskConfiguration, self).__init__(config_string)






#!/usr/bin/env python
# coding=utf-8

"""
Global common functions.
"""

from __future__ import absolute_import
from __future__ import print_function
from lxml import etree
import math
import io
import os
import re
import shutil
import sys
import tempfile
import uuid

import aeneas.globalconstants as gc

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

HHMMSS_MMM_PATTERN = re.compile(r"([0-9]*):([0-9]*):([0-9]*)\.([0-9]*)")
HHMMSS_MMM_PATTERN_COMMA = re.compile(r"([0-9]*):([0-9]*):([0-9]*),([0-9]*)")
PY2 = (sys.version_info[0] == 2)

def uuid_string():
    """
    Return a uuid4 as a Unicode string.

    :rtype: Unicode string
    """
    return safe_unicode(str(uuid.uuid4())).lower()

def custom_tmp_dir():
    """
    Return the path of the temporary directory to use.

    On POSIX OSes (Linux and OS X), return the value of
    :class:`aeneas.globalconstants.RC_TMP_PATH_DEFAULT_POSIX`
    (e.g., ``/tmp/``).

    On non-POSIX OSes, return the value of
    :class:`aeneas.globalconstants.RC_TMP_PATH_DEFAULT_NONPOSIX`
    (i.e., ``None``), so that ``tempfile``
    will use the directory specified by the
    environment/user TMP/TEMP variable.

    :rtype: string (path)
    """
    if is_posix():
        return gc.RC_TMP_PATH_DEFAULT_POSIX
    return gc.RC_TMP_PATH_DEFAULT_NONPOSIX

def tmp_directory(root=None):
    """
    Return the path of a temporary directory created by ``tempfile``.

    :param root: path to the root temporary directory;
                 if ``None``, the default temporary directory
                 will be used instead
    :type  root: Unicode string (path)
    :rtype: string (path)
    """
    if root is None:
        root = custom_tmp_dir()
    return tempfile.mkdtemp(dir=root)

def tmp_file(suffix=u"", root=None):
    """
    Return a (handler, path) tuple
    for a temporary file with given suffix created by ``tempfile``.

    :param suffix: the suffix (e.g., the extension) of the file
    :type  suffix: Unicode string
    :param root: path to the root temporary directory;
                 if ``None``, the default temporary directory
                 will be used instead
    :type  root: Unicode string (path)
    :rtype: tuple
    """
    if root is None:
        root = custom_tmp_dir()
    return tempfile.mkstemp(suffix=suffix, dir=root)

def file_extension(path):
    """
    Return the file extension.

    Examples: ::

        /foo/bar.baz => baz
        None         => None

    :param path: the file path
    :type  path: string (path)
    :rtype: string (path)
    """
    if path is None:
        return None
    ext = os.path.splitext(os.path.basename(path))[1]
    if ext.startswith("."):
        ext = ext[1:]
    return ext

def file_name_without_extension(path):
    """
    Return the file name without extension.

    Examples: ::

        /foo/bar.baz => bar
        /foo/bar     => bar
        None         => None

    :param path: the file path
    :type  path: string (path)
    :rtype: string (path)
    """
    if path is None:
        return None
    return os.path.splitext(os.path.basename(path))[0]

def safe_float(string, default=None):
    """
    Safely parse a string into a float.

    On error return the ``default`` value.

    :param string: string value to be converted
    :type  string: string
    :param default: default value to be used in case of failure
    :type  default: float
    :rtype: float
    """
    value = default
    try:
        value = float(string)
    except TypeError:
        pass
    except ValueError:
        pass
    return value

def safe_int(string, default=None):
    """
    Safely parse a string into an int.

    On error return the ``default`` value.

    :param string: string value to be converted
    :type  string: string
    :param default: default value to be used in case of failure
    :type  default: int
    :rtype: int
    """
    value = safe_float(string, default)
    if value is not None:
        value = int(value)
    return value

def safe_get(dictionary, key, default_value, can_return_none=True):
    """
    Safely perform a dictionary get,
    returning the default value if the key is not found.

    :param dictionary: the dictionary
    :type  dictionary: dict
    :param key: the key
    :type  key: Unicode string
    :param default_value: the default value to be returned
    :type  default_value: variant
    :param can_return_none: if ``True``, the function can return ``None``;
                            otherwise, return ``default_value`` even if the
                            dictionary lookup succeeded
    :type  can_return_none: bool
    :rtype: variant
    """
    return_value = default_value
    try:
        return_value = dictionary[key]
        if (return_value is None) and (not can_return_none):
            return_value = default_value
    except (KeyError, TypeError):
        # KeyError if key is not present in dictionary
        # TypeError if dictionary is None
        pass
    return return_value

def norm_join(prefix, suffix):
    """
    Join ``prefix`` and ``suffix`` paths
    and return the resulting path, normalized.

    :param prefix: the prefix path
    :type  prefix: string (path)
    :param suffix: the suffix path
    :type  suffix: string (path)
    :rtype: string (path)
    """
    if (prefix is None) and (suffix is None):
        return "."
    if prefix is None:
        return os.path.normpath(suffix)
    if suffix is None:
        return os.path.normpath(prefix)
    return os.path.normpath(os.path.join(prefix, suffix))

def config_txt_to_string(string):
    """
    Convert the contents of a TXT config file
    into the corresponding configuration string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    Leading and trailing blank characters will be stripped
    and empty lines (after stripping) will be ignored.

    :param string: the contents of a TXT config file
    :type  string: Unicode string
    :rtype: Unicode string
    """
    if string is None:
        return None
    pairs = [l.strip() for l in string.splitlines() if len(l.strip()) > 0]
    return gc.CONFIG_STRING_SEPARATOR_SYMBOL.join(pairs)

def config_string_to_dict(string, result=None):
    """
    Convert a given configuration string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param string: the configuration string
    :type  string: string
    :rtype: dict
    """
    if string is None:
        return {}
    pairs = string.split(gc.CONFIG_STRING_SEPARATOR_SYMBOL)
    return pairs_to_dict(pairs, result)

def config_xml_to_dict(contents, result, parse_job=True):
    """
    Convert the contents of a XML config file
    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param contents: the XML configuration contents
    :type  contents: bytes
    :param parse_job: if ``True``, parse the job properties;
                      if ``False``, parse the tasks properties
    :type  parse_job: bool
    :rtype: dict (``parse_job=True``) or list of dict (``parse_job=False``)
    """
    try:
        root = etree.fromstring(contents)
        pairs = []
        if parse_job:
            # parse job
            for elem in root:
                if (elem.tag != gc.CONFIG_XML_TASKS_TAG) and (elem.text is not None):
                    pairs.append(u"%s%s%s" % (
                        safe_unicode(elem.tag),
                        gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
                        safe_unicode(elem.text.strip())
                    ))
            return pairs_to_dict(pairs)
        else:
            # parse tasks
            output_list = []
            for task in root.find(gc.CONFIG_XML_TASKS_TAG):
                if task.tag == gc.CONFIG_XML_TASK_TAG:
                    pairs = []
                    for elem in task:
                        if elem.text is not None:
                            pairs.append(u"%s%s%s" % (
                                safe_unicode(elem.tag),
                                gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
                                safe_unicode(elem.text.strip())
                            ))
                    output_list.append(pairs_to_dict(pairs))
            return output_list
    except:
        if result is not None:
            result.passed = False
            result.add_error("An error occurred while parsing XML file")
        if parse_job:
            return {}
        else:
            return []

def config_dict_to_string(dictionary):
    """
    Convert a given config dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    into the corresponding string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    :param dictionary: the config dictionary
    :type  dictionary: dict
    :rtype: string
    """
    parameters = []
    for key in dictionary:
        parameters.append(u"%s%s%s" % (
            key,
            gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
            dictionary[key]
        ))
    return gc.CONFIG_STRING_SEPARATOR_SYMBOL.join(parameters)

def pairs_to_dict(pairs, result=None):
    """
    Convert a given list of ``key=value`` strings ::

        ["key_1=value_1", "key_2=value_2", ..., "key_n=value_n"]

    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param pairs: the list of key=value strings
    :type  pairs: list of strings
    :rtype: dict
    """
    dictionary = {}
    for pair in pairs:
        if len(pair) > 0:
            tokens = pair.split(gc.CONFIG_STRING_ASSIGNMENT_SYMBOL)
            if ((len(tokens) == 2) and
                    (len(tokens[0])) > 0 and
                    (len(tokens[1]) > 0)):
                dictionary[tokens[0]] = tokens[1]
            elif result is not None:
                result.add_warning("Invalid key=value string: '%s'" % pair)
    return dictionary

def copytree(source_directory, destination_directory, ignore=None):
    """
    Recursively copy the contents of a source directory
    into a destination directory.
    Both directories must exist.

    NOTE: this function does not copy the root directory ``source_directory``
    into ``destination_directory``.

    NOTE: ``shutil.copytree(src, dst)`` requires ``dst`` not to exist,
    so we cannot use for our purposes.

    NOTE: code adapted from http://stackoverflow.com/a/12686557

    :param source_directory: the source directory, already existing
    :type  source_directory: string (path)
    :param destination_directory: the destination directory, already existing
    :type  destination_directory: string (path)
    """
    if os.path.isdir(source_directory):
        if not os.path.isdir(destination_directory):
            os.makedirs(destination_directory)
        files = os.listdir(source_directory)
        if ignore is not None:
            ignored = ignore(source_directory, files)
        else:
            ignored = set()
        for f in files:
            if f not in ignored:
                copytree(
                    os.path.join(source_directory, f),
                    os.path.join(destination_directory, f),
                    ignore
                )
    else:
        shutil.copyfile(source_directory, destination_directory)

def ensure_parent_directory(path, ensure_parent=True):
    """
    Ensures the parent directory exists.

    :param path: the path of the file
    :type  path: string (path)
    :param ensure_parent: if True, ensure the parent directory of ``path`` exists;
                       if False, ensure ``path`` exists
    :type  ensure_parent: bool
    :raise OSError: if the path cannot be created
    """
    parent_directory = os.path.abspath(path)
    if ensure_parent:
        parent_directory = os.path.dirname(parent_directory)
    if not os.path.exists(parent_directory):
        try:
            os.makedirs(parent_directory)
        except (IOError, OSError):
            raise OSError("Directory '%s' cannot be created" % parent_directory)

def time_from_ttml(string):
    """
    Parse the given ``SS.mmms`` string
    (TTML values have an "s" suffix, e.g. ``1.234s``)
    and return a float time value.

    :param string: the string to be parsed
    :type  string: unicode
    :rtype: float
    """
    if (string is None) or (len(string) < 2):
        return 0
    # strips "s" at the end
    string = string[:-1]
    return time_from_ssmmm(string)

def time_to_ttml(time_value):
    """
    Format the given time value into a ``SS.mmms`` string
    (TTML values have an "s" suffix, e.g. ``1.234s``).

    Examples: ::

        12        => 12.000s
        12.345    => 12.345s
        12.345432 => 12.345s
        12.345678 => 12.346s

    :param time_value: a time value, in seconds
    :type  time_value: float
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    return "%ss" % time_to_ssmmm(time_value)

def time_from_ssmmm(string):
    """
    Parse the given ``SS.mmm`` string and return a float time value.

    :param string: the string to be parsed
    :type  string: unicode
    :rtype: float
    """
    if (string is None) or (len(string) < 1):
        return 0.000
    return float(string)

def time_to_ssmmm(time_value):
    """
    Format the given time value into a ``SS.mmm`` string.

    Examples: ::

        12        => 12.000
        12.345    => 12.345
        12.345432 => 12.345
        12.345678 => 12.346

    :param time_value: a time value, in seconds
    :type  time_value: float
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    return "%.3f" % (time_value)

def time_from_hhmmssmmm(string, decimal_separator="."):
    """
    Parse the given ``HH:MM:SS.mmm`` string and return a float time value.

    :param string: the string to be parsed
    :type  string: unicode
    :param decimal_separator: the decimal separator to be used
    :type  decimal_separator: string
    :rtype: float
    """
    if decimal_separator == ",":
        pattern = HHMMSS_MMM_PATTERN_COMMA
    else:
        pattern = HHMMSS_MMM_PATTERN
    v_length = 0.000
    try:
        match = pattern.search(string)
        if match is not None:
            v_h = int(match.group(1))
            v_m = int(match.group(2))
            v_s = int(match.group(3))
            v_f = float("0." + match.group(4))
            v_length = v_h * 3600 + v_m * 60 + v_s + v_f
    except:
        pass
    return v_length

def time_to_hhmmssmmm(time_value, decimal_separator="."):
    """
    Format the given time value into a ``HH:MM:SS.mmm`` string.

    Examples: ::

        12        => 00:00:12.000
        12.345    => 00:00:12.345
        12.345432 => 00:00:12.345
        12.345678 => 00:00:12.346
        83        => 00:01:23.000
        83.456    => 00:01:23.456
        83.456789 => 00:01:23.456
        3600      => 01:00:00.000
        3612.345  => 01:00:12.345

    :param time_value: a time value, in seconds
    :type  time_value: float
    :param decimal_separator: the decimal separator, default ``.``
    :type  decimal_separator: char
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    tmp = time_value
    hours = int(math.floor(tmp / 3600))
    tmp -= (hours * 3600)
    minutes = int(math.floor(tmp / 60))
    tmp -= minutes * 60
    seconds = int(math.floor(tmp))
    tmp -= seconds
    milliseconds = int(math.floor(tmp * 1000))
    return "%02d:%02d:%02d%s%03d" % (
        hours,
        minutes,
        seconds,
        decimal_separator,
        milliseconds
    )

def time_to_srt(time_value):
    """
    Format the given time value into a ``HH:MM:SS,mmm`` string,
    as used in the SRT format.

    Examples: ::

        12        => 00:00:12,000
        12.345    => 00:00:12,345
        12.345432 => 00:00:12,345
        12.345678 => 00:00:12,346
        83        => 00:01:23,000
        83.456    => 00:01:23,456
        83.456789 => 00:01:23,456
        3600      => 01:00:00,000
        3612.345  => 01:00:12,345

    :param time_value: a time value, in seconds
    :type  time_value: float
    :rtype: string
    """
    return time_to_hhmmssmmm(time_value, ",")

def split_url(url):
    """
    Split the given URL base#anchor into (base, anchor),
    or (base, None) if no anchor is present.

    :param url: the url
    :type  url: str
    :rtype: list of str
    """
    if url is None:
        return (None, None)
    array = url.split("#")
    if len(array) == 1:
        array.append(None)
    elif len(array) > 2:
        # TODO raise an exception?
        array = array[0:2]
    return tuple(array)

def is_posix():
    """
    Return ``True`` if running on a POSIX OS.

    :rtype: bool
    """
    # from https://docs.python.org/2/library/os.html#os.name
    # the registered values of os.name are:
    # "posix", "nt", "os2", "ce", "java", "riscos"
    return os.name == "posix"

def is_linux():
    """
    Return ``True`` if running on Linux.

    :rtype: bool
    """
    return (is_posix()) and (os.uname()[0] == "Linux")

def is_osx():
    """
    Return ``True`` if running on Mac OS X (Darwin).

    :rtype: bool
    """
    return (is_posix()) and (os.uname()[0] == "Darwin")

def is_windows():
    """
    Return ``True`` if running on Windows.

    :rtype: bool
    """
    return os.name == "nt"

def fix_slash(path):
    """
    On non-POSIX OSes, change the slashes in ``path``
    for loading in the browser.

    Example: ::

        c:\\abc\\def => c:/abc/def

    :param path: the path
    :type  path: Unicode string (path)
    :rtype: Unicode string
    """
    if not is_posix():
        # TODO is there a better way to do this?
        path = path.replace("\\", "/")
    return path

def can_run_c_extension(name=None):
    """
    Determine whether the given Python C extension loads correctly.

    If ``name`` is ``None``, tests all Python C extensions,
    and return ``True`` if and only if all load correctly.

    :param name: the name of the Python C extension to test
    :type  name: string
    :rtype: bool
    """
    def can_run_cdtw():
        """ Python C extension for computing DTW """
        try:
            import aeneas.cdtw.cdtw
            return True
        except ImportError:
            return False
    def can_run_cmfcc():
        """ Python C extension for computing MFCC """
        try:
            import aeneas.cmfcc.cmfcc
            return True
        except ImportError:
            return False
    def can_run_cew():
        """ Python C extension for synthesizing with espeak """
        try:
            import aeneas.cew.cew
            return True
        except ImportError:
            return False

    if name == "cdtw":
        return can_run_cdtw()
    elif name == "cmfcc":
        return can_run_cmfcc()
    elif name == "cew":
        return can_run_cew()
    else:
        if is_linux():
            # Linux
            return can_run_cdtw() and can_run_cmfcc() and can_run_cew()
        else:
            # no cew for other OSes
            return can_run_cdtw() and can_run_cmfcc()

def run_c_extension_with_fallback(
        log_function,
        extension,
        c_function,
        py_function,
        args,
        c_extension=True
):
    """
    Run a function calling a C extension, falling back
    to a pure Python function if the former does not succeed.

    :param log_function: a logger function (see ``_log()`` in many classes)
    :type  log_function: function
    :param extension: the name of the extension
    :type  extension: string
    :param c_function: the (Python) function calling the C extension
    :type  c_function: function
    :param py_function: the (Python) function providing the fallback
    :type  py_function: function
    :param c_extension: if ``True``, try running the C extension first;
                        if ``False``, directly run the pure Python fallback
    :type  c_extension: bool
    :rtype: depends on the extension being called

    :raise RuntimeError: if both the C extension and
                         the pure Python code did not succeed.

    .. versionadded:: 1.4.0
    """
    computed = False
    if c_extension:
        log_function(u"C extensions enabled")
        if can_run_c_extension(extension):
            log_function([u"C extensions enabled and %s can be loaded", extension])
            computed, result = c_function(*args)
        else:
            log_function([u"C extensions enabled but %s cannot be loaded", extension])
    else:
        log_function(u"C extensions disabled")
    if not computed:
        log_function(u"Running the pure Python code")
        computed, result = py_function(*args)
    if not computed:
        # TODO create a more meaningful message
        raise RuntimeError("Both the C extension and the pure Python code failed. (Wrong arguments? Input too big?)")
    return result

def file_can_be_read(path):
    """
    Return ``True`` if the file at the given ``path`` can be read.

    :param path: the file path
    :type  path: string (path)
    :rtype: bool

    .. versionadded:: 1.4.0
    """
    if path is None:
        return False
    try:
        # TODO is doing this with os attributes preferable?
        test_file = io.open(path, "rb")
        test_file.close()
        return True
    except (IOError, OSError):
        pass
    return False

def file_can_be_written(path):
    """
    Return ``True`` if a file can be written at the given ``path``.

    IMPORTANT: this function will attempt to open the given ``path``
    in write mode, possibly destroying the file previously existing there.

    :param path: the file path
    :type  path: string (path)
    :rtype: bool

    .. versionadded:: 1.4.0
    """
    if path is None:
        return False
    try:
        # TODO think if it is better doing this with os attributes
        test_file = io.open(path, "wb")
        test_file.close()
        delete_file(None, path)
        return True
    except (IOError, OSError):
        pass
    return False

def directory_exists(path):
    """
    Return ``True`` if the given ``path`` string
    points to an existing directory.

    :param path: the file path
    :type  path: string (path)
    :rtype: bool
    """
    if (path is None) or (not os.path.isdir(path)):
        return False
    return True

def file_exists(path):
    """
    Return ``True`` if the given ``path`` string
    points to an existing file.

    :param path: the file path
    :type  path: string (path)
    :rtype: bool
    """
    if (path is None) or (not os.path.isfile(path)):
        return False
    return True

def file_size(path):
    """
    Return the size, in bytes, of the file at the given ``path``.
    Return ``-1`` if the file does not exist or cannot be read.

    :param path: the file path
    :type  path: string (path)
    :rtype: int
    """
    try:
        return os.path.getsize(path)
    except OSError:
        return -1

def delete_directory(path):
    """
    Safely delete a directory.

    :param path: the file path
    :type  path: string (path)
    """
    if path is not None:
        try:
            shutil.rmtree(path)
        except:
            pass

def delete_file(handler, path):
    """
    Safely delete file.

    :param handler: the file handler (as returned by tempfile)
    :type  handler: obj
    :param path: the file path
    :type  path: string (path)
    """
    if handler is not None:
        try:
            os.close(handler)
        except:
            pass
    if path is not None:
        try:
            os.remove(path)
        except:
            pass

def relative_path(path, from_file):
    """
    Return the relative path of a file or directory, specified
    as ``path`` relative to (the parent directory of) ``from file``.

    The returned path is relative to the current working directory.

    Example: ::

        path="res/foo.bar"
        from_file="/root/abc/def/ghi.py"
        cwd="/root"
        => "abc/def/res/foo.bar"

    :param path: the file path
    :type  path: string (path)
    :param from_file: the reference file
    :type  from_file: string (path)
    :rtype: string (path)
    """
    if path is None:
        return None
    abs_path_target = absolute_path(path, from_file)
    abs_path_cwd = os.getcwd()
    return os.path.relpath(abs_path_target, start=abs_path_cwd)

def absolute_path(path, from_file):
    """
    Return the absolute path of a file or directory, specified
    as ``path`` relative to (the parent directory of) ``from_file``.

    This method is intented to be called with ``__file__``
    as second argument.

    Example: ::

        path="res/foo.bar"
        from_file="/abc/def/ghi.py"
        => "/abc/def/res/foo.bar"

    :param path: the file path
    :type  path: string (path)
    :param from_file: the reference file
    :type  from_file: string (path)
    :rtype: string (path)
    """
    if path is None:
        return None
    current_directory = os.path.dirname(from_file)
    target = os.path.join(current_directory, path)
    return os.path.abspath(target)

def read_file_bytes(input_file_path):
    """
    Read the file at the given file path
    and return its contents as a byte string,
    or ``None`` if an error occurred.

    :param input_file_path: the file path
    :type  input_file_path: string (path)
    :rtype: byte string
    """
    contents = None
    try:
        with io.open(input_file_path, "rb") as input_file:
            contents = input_file.read()
    except:
        pass
    return contents

def human_readable_number(number, suffix=""):
    """
    Format the given number into a human-readable string.

    :param number: the number
    :type  number: int or float
    :param suffix: the unit of the number
    :type  suffix: str

    Code adapted from http://stackoverflow.com/a/1094933
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(number) < 1024.0:
            return "%3.1f%s%s" % (number, unit, suffix)
        number /= 1024.0
    return "%.1f%s%s" % (number, "Y", suffix)

def is_unicode(string):
    """
    Return True if the given string is a sequence of Unicode code points.

    :param string: the string to test
    :type  string: any
    :rtype: bool
    """
    if PY2:
        return isinstance(string, unicode)
    return isinstance(string, str)

def is_bytes(string):
    """
    Return True if the given string is a sequence of bytes.

    :param string: the string to test
    :type  string: any
    :rtype: bool
    """
    if PY2:
        return isinstance(string, str)
    return isinstance(string, bytes)

def is_utf8_encoded(bstring):
    """
    Return True if the given byte string can be decoded
    into a Unicode string using the UTF-8 decoder.

    :param bstring: the string to test
    :type  bstring: byte string
    :rtype: bool
    """
    try:
        bstring.decode("utf-8")
        return True
    except UnicodeDecodeError:
        pass
    return False

def safe_str(string):
    """
    Safely return the given Unicode string
    from a __str__ function: as a byte string
    in Python 2, or as a Unicode string in Python 3.

    :param string: the string to return
    :type  string: Unicode string
    :rtype: byte string or Unicode string
    """
    if string is None:
        return None
    if PY2:
        return string.encode("utf-8")
    return string

def safe_unichr(codepoint):
    """
    Safely return a Unicode string of length one,
    containing the Unicode character with given codepoint.

    :param codepoint: the codepoint
    :type  codepoint: int
    :rtype: Unicode string
    """
    if PY2:
        return unichr(codepoint)
    return chr(codepoint)

def safe_unicode(string):
    """
    Safely convert the given string to a Unicode string.

    :param string: the string to convert
    :type  string: byte string or Unicode string
    :rtype: Unicode string
    """
    if string is None:
        return None
    if is_bytes(string):
        return string.decode("utf-8")
    return string

def safe_bytes(string):
    """
    Safely convert the given string to a bytes string.

    :param string: the string to convert
    :type  string: byte string or Unicode string
    :rtype: byte string
    """
    if string is None:
        return None
    if is_unicode(string):
        return string.encode("utf-8")
    return string

def safe_unicode_stdin(string):
    """
    Safely convert the given string to a Unicode string,
    decoding using ``sys.stdin.encoding`` if needed.

    :param string: the string to convert
    :type  string: byte string or Unicode string
    :rtype: Unicode string
    """
    if string is None:
        return None
    if is_bytes(string):
        try:
            return string.decode(sys.stdin.encoding)
        except UnicodeDecodeError:
            return string.decode(sys.stdin.encoding, "replace")
    return string

def safe_print(string):
    """
    Safely print a given Unicode string to stdout,
    possibly replacing characters non-printable
    in the current stdout encoding.

    :param string: the message string
    :type  string: Unicode string
    """
    try:
        print(string)
    except UnicodeEncodeError:
        try:
            # NOTE encoding and decoding so that in Python 3 no b"..." is printed
            encoded = string.encode(sys.stdout.encoding, "replace")
            decoded = encoded.decode(sys.stdout.encoding, "replace")
            print(decoded)
        except (UnicodeDecodeError, UnicodeEncodeError):
            print(u"[ERRO] An unexpected error happened while printing to stdout.")
            print(u"[ERRO] Please check that your file/string encoding matches the shell encoding.")
            print(u"[ERRO] If possible, set your shell encoding to UTF-8 and convert any files with legacy encodings.")

def object_to_unicode(obj):
    """
    Return a sequence of Unicode code points from the given object.

    :param obj: the object
    :type  obj: object
    :rtype: unicode
    """
    if PY2:
        return unicode(obj)
    return str(obj)

def object_to_bytes(obj):
    """
    Return a sequence of bytes from the given object.

    :param obj: the object
    :type  obj: object
    :rtype: unicode
    """
    if PY2:
        return str(obj)
    return bytes(obj, encoding="utf-8")






#!/usr/bin/env python
# coding=utf-8

"""
A class representing an audio file.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy

from aeneas.ffprobewrapper import FFPROBEParsingError
from aeneas.ffprobewrapper import FFPROBEPathError
from aeneas.ffprobewrapper import FFPROBEUnsupportedFormatError
from aeneas.ffprobewrapper import FFPROBEWrapper
from aeneas.logger import Logger
from aeneas.mfcc import MFCC
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.wavfile import read as scipywavread
from aeneas.wavfile import write as scipywavwrite
import aeneas.globalfunctions as gf

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class AudioFileProbeError(Exception):
    """
    Error raised when the probe executable cannot be executed.
    """
    pass



class AudioFileUnsupportedFormatError(Exception):
    """
    Error raised when the format of the given file cannot be decoded.
    """
    pass



class AudioFile(object):
    """
    A class representing an audio file.

    The properties of the audio file (length, format, etc.)
    are set by invoking the ``read_properties()`` function,
    which calls an audio file probe.
    (Currently, the probe is :class:`aeneas.ffprobewrapper.FFPROBEWrapper`)

    :param file_path: the path of the audio file
    :type  file_path: Unicode string (path)
    :param rconf: a runtime configuration. Default: ``None``, meaning that
                  default settings will be used.
    :type  rconf: :class:`aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`aeneas.logger.Logger`
    """

    TAG = u"AudioFile"

    def __init__(self, file_path=None, rconf=None, logger=None):
        self.logger = logger or Logger()
        self.rconf = rconf or RuntimeConfiguration()
        self.file_path = file_path
        self.file_size = None
        self.audio_length = None
        self.audio_format = None
        self.audio_sample_rate = None
        self.audio_channels = None

    def _log(self, message, severity=Logger.DEBUG):
        """ Log """
        self.logger.log(message, severity, self.TAG)

    def __unicode__(self):
        msg = [
            u"File path:         %s" % self.file_path,
            u"File size (bytes): %s" % gf.safe_int(self.file_size),
            u"Audio length (s):  %s" % gf.safe_float(self.audio_length),
            u"Audio format:      %s" % self.audio_format,
            u"Audio sample rate: %s" % gf.safe_int(self.audio_sample_rate),
            u"Audio channels:    %s" % gf.safe_int(self.audio_channels),
        ]
        return u"\n".join(msg)

    def __str__(self):
        return gf.safe_str(self.__unicode__())

    @property
    def file_path(self):
        """
        The path of the audio file.

        :rtype: Unicode string
        """
        return self.__file_path
    @file_path.setter
    def file_path(self, file_path):
        self.__file_path = file_path

    @property
    def file_size(self):
        """
        The size of the audio file, in bytes.

        :rtype: int
        """
        return self.__file_size
    @file_size.setter
    def file_size(self, file_size):
        self.__file_size = file_size

    @property
    def audio_length(self):
        """
        The length of the audio file, in seconds.

        :rtype: float
        """
        return self.__audio_length
    @audio_length.setter
    def audio_length(self, audio_length):
        self.__audio_length = audio_length

    @property
    def audio_format(self):
        """
        The format of the audio file.

        :rtype: Unicode string
        """
        return self.__audio_format
    @audio_format.setter
    def audio_format(self, audio_format):
        self.__audio_format = audio_format

    @property
    def audio_sample_rate(self):
        """
        The sample rate of the audio file, in samples per second.

        :rtype: int
        """
        return self.__audio_sample_rate
    @audio_sample_rate.setter
    def audio_sample_rate(self, audio_sample_rate):
        self.__audio_sample_rate = audio_sample_rate

    @property
    def audio_channels(self):
        """
        The number of channels of the audio file.

        :rtype: int
        """
        return self.__audio_channels
    @audio_channels.setter
    def audio_channels(self, audio_channels):
        self.__audio_channels = audio_channels

    def read_properties(self):
        """
        Populate this object by reading
        the audio properties of the file at the given path.

        Currently this function uses
        :class:`aeneas.ffprobewrapper.FFPROBEWrapper`
        to get the audio file properties.

        :raises AudioFileProbeError: if the path to the ``ffprobe`` executable cannot be called
        :raises AudioFileUnsupportedFormatError: if the audio file has a format not supported
        :raises OSError: if the audio file cannot be read
        """

        self._log(u"Reading properties...")

        # check the file can be read
        if not gf.file_can_be_read(self.file_path):
            self._log([u"File '%s' cannot be read", self.file_path], Logger.CRITICAL)
            raise OSError(u"File '%s' cannot be read" % self.file_path)

        # get the file size
        self._log([u"Getting file size for '%s'", self.file_path])
        self.file_size = gf.file_size(self.file_path)
        self._log([u"File size for '%s' is '%d'", self.file_path, self.file_size])

        # get the audio properties using FFPROBEWrapper
        try:
            self._log(u"Reading properties with FFPROBEWrapper...")
            properties = FFPROBEWrapper(rconf=self.rconf, logger=self.logger).read_properties(self.file_path)
            self._log(u"Reading properties with FFPROBEWrapper... done")
        except FFPROBEPathError:
            self._log(u"Reading properties with FFPROBEWrapper... failed", Logger.CRITICAL)
            self._log(u"Unable to call ffprobe executable", Logger.CRITICAL)
            raise AudioFileProbeError("Unable to call the audio probe executable")
        except FFPROBEUnsupportedFormatError:
            self._log(u"Reading properties with FFPROBEWrapper... failed", Logger.CRITICAL)
            self._log(u"Unsupported audio file format", Logger.CRITICAL)
            raise AudioFileUnsupportedFormatError("Unsupported audio file format")
        except FFPROBEParsingError:
            self._log(u"Reading properties with FFPROBEWrapper... failed", Logger.CRITICAL)
            self._log(u"Failed while parsing the ffprobe output", Logger.CRITICAL)
            raise AudioFileUnsupportedFormatError("Unsupported audio file format")

        # save relevant properties in results inside the audiofile object
        self.audio_length = gf.safe_float(properties[FFPROBEWrapper.STDOUT_DURATION])
        self.audio_format = properties[FFPROBEWrapper.STDOUT_CODEC_NAME]
        self.audio_sample_rate = gf.safe_int(properties[FFPROBEWrapper.STDOUT_SAMPLE_RATE])
        self.audio_channels = gf.safe_int(properties[FFPROBEWrapper.STDOUT_CHANNELS])
        self._log([u"Stored audio_length: '%s'", self.audio_length])
        self._log([u"Stored audio_format: '%s'", self.audio_format])
        self._log([u"Stored audio_sample_rate: '%s'", self.audio_sample_rate])
        self._log([u"Stored audio_channels: '%s'", self.audio_channels])
        self._log(u"Reading properties... done")



class AudioFileMonoWAVE(AudioFile):
    """
    A monoaural (single-channel) WAVE audio file.

    Its data can be read from and write to file, set from a ``numpy`` 1D array.

    It supports append, prepend, reverse, and trim operations.

    It can also extract MFCCs and store them internally,
    also after the audio data has been discarded.

    NOTE
    At the moment, the state of this object might be inconsistent
    (e.g., setting a new path after loading audio data will not flush the audio data).
    Use this class with care.

    :param file_path: the path of the audio file
    :type  file_path: Unicode string (path)
    :param rconf: a runtime configuration. Default: ``None``, meaning that
                  default settings will be used.
    :type  rconf: :class:`aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`aeneas.logger.Logger`
    """

    TAG = u"AudioFileMonoWAVE"

    def __init__(self, file_path=None, rconf=None, logger=None):
        self.logger = logger or Logger()
        self.rconf = rconf or RuntimeConfiguration()
        self.audio_data = None
        self.audio_mfcc = None
        AudioFile.__init__(self, file_path=file_path, rconf=rconf, logger=logger)

    @property
    def audio_data(self):
        """
        The audio data.

        :rtype: numpy 1D array
        """
        return self.__audio_data
    @audio_data.setter
    def audio_data(self, audio_data):
        self.__audio_data = audio_data

    @property
    def audio_mfcc(self):
        """
        The MFCCs of the audio file.

        :rtype: numpy 2D array
        """
        return self.__audio_mfcc

    @audio_mfcc.setter
    def audio_mfcc(self, audio_mfcc):
        self.__audio_mfcc = audio_mfcc

    def load_data(self):
        """
        Load the audio file data.

        :raises AudioFileUnsupportedFormatError: if the audio file is not a mono WAVE file
        :raises OSError: if the audio file cannot be read
        """
        self._log(u"Loading audio data...")

        # check the file can be read
        if not gf.file_can_be_read(self.file_path):
            self._log([u"File '%s' cannot be read", self.file_path], Logger.CRITICAL)
            raise OSError("File '%s' cannot be read" % self.file_path)

        try:
            self.audio_format = "pcm16"
            self.audio_sample_rate, self.audio_data = scipywavread(self.file_path)
            # scipy reads a sample as an int16_t, that is, a number in [-32768, 32767]
            # so we convert it to a float64 in [-1, 1]
            self.audio_data = self.audio_data.astype("float64") / 32768
        except ValueError:
            self._log(u"Unsupported audio file format", Logger.CRITICAL)
            raise AudioFileUnsupportedFormatError("Unsupported audio file format")

        self._update_length()
        self._log([u"Sample length: %f", self.audio_length])
        self._log([u"Sample rate:   %f", self.audio_sample_rate])
        self._log([u"Audio format:  %s", self.audio_format])
        self._log(u"Loading audio data... done")

    def append_data(self, new_data):
        """
        Append the given new data to the current audio data.

        If audio data is not loaded, create an empty data structure
        and then append to it.

        :param new_data: the new data to be appended
        :type  new_data: numpy 1D array

        .. versionadded:: 1.2.1
        """
        self._log(u"Appending audio data...")
        self._audio_data_is_initialized(load=False)
        self.audio_data = numpy.append(self.audio_data, new_data)
        self._update_length()
        self._log(u"Appending audio data... done")

    def prepend_data(self, new_data):
        """
        Prepend the given new data to the current audio data.

        If audio data is not loaded, create an empty data structure
        and then preppend to it.

        :param new_data: the new data to be prepended
        :type  new_data: numpy 1D array

        .. versionadded:: 1.2.1
        """
        self._log(u"Prepending audio data...")
        self._audio_data_is_initialized(load=False)
        self.audio_data = numpy.append(new_data, self.audio_data)
        self._update_length()
        self._log(u"Prepending audio data... done")

    def extract_mfcc(self):
        """
        Extract MFCCs from the given audio file.

        If audio data is not loaded, load it, extract MFCCs,
        store them internally, and discard the audio data immediately.

        :raise RuntimeError: if both the C extension and
                             the pure Python code did not succeed.
        """
        had_audio_data = self._audio_data_is_initialized(load=True)
        gf.run_c_extension_with_fallback(
            self._log,
            "cmfcc",
            self._compute_mfcc_c_extension,
            self._compute_mfcc_pure_python,
            (),
            c_extension=self.rconf["c_ext"]
        )
        if not had_audio_data:
            self._log(u"Audio data was not loaded, clearing it")
            self.clear_data()
        else:
            self._log(u"Audio data was loaded, not clearing it")

    def reverse(self):
        """
        Reverse the audio data.

        If audio data is not loaded, load it and then reverse it.

        .. versionadded:: 1.2.0
        """
        self._log(u"Reversing...")
        self._audio_data_is_initialized(load=True)
        self.audio_data = self.audio_data[::-1]
        self._log(u"Reversing... done")

    def trim(self, begin=None, length=None):
        """
        Get a slice of the audio data of ``length`` seconds,
        starting from ``begin`` seconds.

        If audio data is not loaded, load it and then slice it.

        :param begin: the start position, in seconds
        :type  begin: float
        :param length: the  position, in seconds
        :type  length: float

        .. versionadded:: 1.2.0
        """
        self._log(u"Trimming...")
        if (begin is None) and (length is None):
            self._log(u"begin and length are both None: nothing to do")
        else:
            self._audio_data_is_initialized(load=True)
            self._log([u"audio_length is %.3f", self.audio_length])
            if begin is None:
                begin = 0
                self._log([u"begin was None, now set to %.3f", begin])
            begin = min(max(0, begin), self.audio_length)
            self._log([u"begin is %.3f", begin])
            if length is None:
                length = self.audio_length - begin
                self._log([u"length was None, now set to %.3f", length])
            length = min(max(0, length), self.audio_length - begin)
            self._log([u"length is %.3f", length])
            begin_index = int(begin * self.audio_sample_rate)
            end_index = int((begin + length) * self.audio_sample_rate)
            self.audio_data = self.audio_data[begin_index:end_index]
            self._update_length()
        self._log(u"Trimming... done")

    def write(self, file_path):
        """
        Write the audio data to file.
        Return ``True`` on success, or ``False`` otherwise.

        :param file_path: the path of the output file to be written
        :type  file_path: Unicode string (path)

        .. versionadded:: 1.2.0
        """
        self._log([u"Writing audio file '%s'...", file_path])
        self._audio_data_is_initialized(load=False)
        try:
            # our value is a float64 in [-1, 1]
            # scipy writes the sample as an int16_t, that is, a number in [-32768, 32767]
            data = (self.audio_data * 32768).astype("int16")
            scipywavwrite(file_path, self.audio_sample_rate, data)
        except:
            self._log(u"Error writing audio file", severity=Logger.CRITICAL)
            raise OSError("Error writing audio file to '%s'" % file_path)
        self._log([u"Writing audio file '%s'... done", file_path])

    def clear_data(self):
        """
        Clear the audio data, freeing memory.
        """
        self._log(u"Clear audio_data")
        self.audio_data = None

    def _update_length(self):
        """
        Update the audio length property,
        according to the length of the current audio data
        and audio sample rate.

        This function fails silently if one of the two is None.
        """
        if (self.audio_sample_rate is not None) and (self.audio_data is not None):
            self.audio_length = len(self.audio_data) / self.audio_sample_rate

    def _audio_data_is_initialized(self, load=True):
        """
        Check if audio data is loaded:
        if so, return True.

        Otherwise, either load or initialize the audio data
        and return False.

        :param load: if True, load from file; if False, initialize to empty
        :type  load: bool
        :rtype: bool
        """
        if self.audio_data is not None:
            self._log(u"audio data is not None: returning True")
            return True
        if load:
            self._log(u"No audio data: loading it from file")
            self.load_data()
        else:
            self._log(u"No audio data: initializing it to an empty data structure")
            self.audio_data = numpy.array([])
        self._log(u"audio data was None: returning False")
        return False

    def _compute_mfcc_c_extension(self):
        """
        Compute MFCCs using the Python C extension cmfcc.
        """
        self._log(u"Computing MFCCs using C extension...")
        try:
            self._log(u"Importing cmfcc...")
            import aeneas.cmfcc.cmfcc
            self._log(u"Importing cmfcc... done")
            self.audio_mfcc = (aeneas.cmfcc.cmfcc.compute_from_data(
                self.audio_data,
                self.audio_sample_rate,
                self.rconf["mfcc_filters"],
                self.rconf["mfcc_size"],
                self.rconf["mfcc_order"],
                self.rconf["mfcc_lower_freq"],
                self.rconf["mfcc_upper_freq"],
                self.rconf["mfcc_emph"],
                self.rconf["mfcc_win_len"],
                self.rconf["mfcc_win_shift"]
            )[0]).transpose()
            self._log(u"Computing MFCCs using C extension... done")
            return (True, None)
        except Exception as exc:
            self._log(u"Computing MFCCs using C extension... failed")
            self._log(u"An unexpected exception occurred while running cmfcc:", Logger.WARNING)
            self._log([u"%s", exc], Logger.WARNING)
        return (False, None)

    def _compute_mfcc_pure_python(self):
        """
        Compute MFCCs using the pure Python code.
        """
        self._log(u"Computing MFCCs using pure Python code...")
        try:
            self.audio_mfcc = MFCC(
                rconf=self.rconf,
                logger=self.logger
            ).compute_from_data(self.audio_data, self.audio_sample_rate).transpose()
            self._log(u"Computing MFCCs using pure Python code... done")
            return (True, None)
        except Exception as exc:
            self._log(u"Computing MFCCs using pure Python code... failed")
            self._log(u"An unexpected exception occurred while running pure Python code:", Logger.WARNING)
            self._log([u"%s", exc], Logger.WARNING)
        return (False, None)






#!/usr/bin/env python
# coding=utf-8

"""
Set aeneas package up
"""

from setuptools import setup, Extension
import io
import os
import sys

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL 3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

try:
    from numpy import get_include
    from numpy.distutils import misc_util
except ImportError:
    print("[ERRO] You must install numpy before installing aeneas")
    print("[INFO] Try the following command:")
    print("[INFO] $ sudo pip install numpy")
    sys.exit(1)

IS_LINUX = (os.name == "posix") and (os.uname()[0] == "Linux")

SHORT_DESCRIPTION = "aeneas is a Python/C library and a set of tools to automagically synchronize audio and text (aka forced alignment)"

LONG_DESCRIPTION = io.open("README.rst", "r", encoding="utf-8").read()

INCLUDE_DIRS = [misc_util.get_numpy_include_dirs()]

EXTENSION_CDTW = Extension(
    "aeneas.cdtw.cdtw",
    ["aeneas/cdtw/cdtw_py.c", "aeneas/cdtw/cdtw_func.c"],
    include_dirs=[get_include()]
)
EXTENSION_CEW = Extension(
    "aeneas.cew.cew",
    ["aeneas/cew/cew_py.c", "aeneas/cew/cew_func.c"],
    libraries=["espeak"]
)
EXTENSION_CMFCC = Extension(
    "aeneas.cmfcc.cmfcc",
    ["aeneas/cmfcc/cmfcc_py.c", "aeneas/cmfcc/cmfcc_func.c", "aeneas/cmfcc/cwave_func.c"],
    include_dirs=[get_include()]
)
# cwave is ready, but currently not used
#EXTENSION_CWAVE = Extension(
#    "aeneas.cwave.cwave",
#    ["aeneas/cwave/cwave_py.c", "aeneas/cwave/cwave_func.c"],
#    include_dirs=[get_include()]
#)
#EXTENSIONS = [EXTENSION_CDTW, EXTENSION_CMFCC, EXTENSION_CWAVE]

EXTENSIONS = [EXTENSION_CDTW, EXTENSION_CMFCC]
if IS_LINUX:
    # cew is available only for Linux at the moment
    EXTENSIONS.append(EXTENSION_CEW)

setup(
    name="aeneas",
    packages=[
        "aeneas",
        "aeneas.cdtw",
        "aeneas.cew",
        "aeneas.cmfcc",
        "aeneas.cwave",
        "aeneas.tools"
    ],
    package_data={
        "aeneas": ["res/*"],
        "aeneas.cdtw": ["*.c", "*.h"],
        "aeneas.cew": ["*.c", "*.h"],
        "aeneas.cmfcc": ["*.c", "*.h"],
        "aeneas.cwave": ["*.c", "*.h"],
        "aeneas.tools": ["res/*"]
    },
    version="1.4.1.0",
    description=SHORT_DESCRIPTION,
    author="Alberto Pettarin",
    author_email="alberto@albertopettarin.it",
    url="https://github.com/readbeyond/aeneas",
    license="GNU Affero General Public License v3 (AGPL v3)",
    long_description=LONG_DESCRIPTION,
    install_requires=[
        "BeautifulSoup4>=4.4",
        "lxml>=3.0",
        "numpy>=1.9"
    ],
    extras_require={"pafy": ["pafy>=0.3"]},
    scripts=[
        "bin/aeneas_convert_syncmap",
        "bin/aeneas_download",
        "bin/aeneas_execute_job",
        "bin/aeneas_execute_task",
        "bin/aeneas_synthesize_text",
        "bin/aeneas_validate",
    ],
    keywords=[
        "CSV",
        "DTW",
        "EPUB 3 Media Overlay",
        "EPUB 3",
        "EPUB",
        "JSON",
        "MFCC",
        "Mel-frequency cepstral coefficients",
        "ReadBeyond Sync",
        "ReadBeyond",
        "SMIL",
        "SRT",
        "SSV",
        "TSV",
        "TTML",
        "VTT",
        "XML",
        "aeneas",
        "audio/text alignment",
        "dynamic time warping",
        "espeak",
        "ffmpeg",
        "ffprobe",
        "forced alignment",
        "media overlay",
        "rb_smil_emulator",
        "subtitles",
        "sync",
        "synchronization",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: C",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Education",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Topic :: Multimedia :: Sound/Audio :: Speech",
        "Topic :: Printing",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Utilities"
    ],
    ext_modules=EXTENSIONS,
    include_dirs=INCLUDE_DIRS,
)



#!/usr/bin/env python
# coding=utf-8

"""
This module contains the implementation
of a simple Start Detector (SD),
based on VAD and iterated DTW.

Given a (full) audio file and the corresponding (full) text,
it will compute the time interval
containing the given text,
that is, detect the audio head and the audio length.

.. versionadded:: 1.2.0
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy

from aeneas.audiofile import AudioFileMonoWAVE
from aeneas.dtw import DTWAligner
from aeneas.logger import Logger
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.synthesizer import Synthesizer
from aeneas.vad import VAD
import aeneas.globalfunctions as gf

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class SDMetric(object):
    """
    Enumeration of the SD metrics that can be used
    for selecting the best SD candidate.
    """

    VALUE = "value"
    """ Use the value from the accumulated cost matrix """

    DISTORTION = "distortion"
    """ Use the distortion (value / length) """


class SD(object):
    """
    The SD extractor.

    :param audio_file: the audio file
    :type  audio_file: :class:`aeneas.audiofile.AudioFileMonoWAVE`
    :param text_file: the text file
    :type  text_file: :class:`aeneas.textfile.TextFile`
    :param rconf: a runtime configuration. Default: ``None``, meaning that
                  default settings will be used.
    :type  rconf: :class:`aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`aeneas.logger.Logger`
    """

    TAG = u"SD"

    # TODO eliminate these magic numbers
    MAX_RUNS_NO_IMPROVEMENT = 20
    MAX_RUNS_WITH_MIN_LENGTH = 20
    QUERY_FACTOR = 2.0
    AUDIO_FACTOR = 6.0

    MAX_HEAD_LENGTH = 10.0
    """
    Try detecting audio heads up to this many seconds.
    Default: ``10.0``.

    .. versionadded:: 1.2.0
    """

    MIN_HEAD_LENGTH = 0.0
    """
    Try detecting audio heads of at least this many seconds.
    Default: ``0.0``.

    .. versionadded:: 1.2.0
    """

    MAX_TAIL_LENGTH = 10.0
    """
    Try detecting audio tails up to this many seconds.
    Default: ``10.0``.

    .. versionadded:: 1.2.0
    """

    MIN_TAIL_LENGTH = 0.0
    """
    Try detecting audio tails of at least this many seconds.
    Default: ``0.0``.

    .. versionadded:: 1.2.0
    """

    def __init__(self, audio_file, text_file, rconf=None, logger=None):
        self.logger = logger or Logger()
        self.rconf = rconf or RuntimeConfiguration()
        self.audio_file = audio_file
        self.text_file = text_file
        self.audio_speech = None

    def _log(self, message, severity=Logger.DEBUG):
        """ Log """
        self.logger.log(message, severity, self.TAG)

    def detect_interval(
            self,
            min_head_length=MIN_HEAD_LENGTH,
            max_head_length=MAX_HEAD_LENGTH,
            min_tail_length=MIN_TAIL_LENGTH,
            max_tail_length=MAX_TAIL_LENGTH,
            metric=SDMetric.VALUE
        ):
        """
        Detect the audio interval.

        :param max_head_length: estimated maximum head length
        :type  max_head_length: float
        :param max_tail_length: estimated maximum tail length
        :type  max_tail_length: float
        :param metric: the metric to be used when comparing candidates
        :type  metric: :class:`aeneas.sd.SDMetric`
        :rtype: (float, float)
        """
        head = self.detect_head(min_head_length, max_head_length, metric)
        tail = self.detect_tail(min_tail_length, max_tail_length, metric)
        begin = head
        end = self.audio_file.audio_length - tail
        self._log([u"Audio length: %.3f", self.audio_file.audio_length])
        self._log([u"Head length:  %.3f", head])
        self._log([u"Tail length:  %.3f", tail])
        self._log([u"Begin:        %.3f", begin])
        self._log([u"End:          %.3f", end])
        if (begin >= 0) and (end > begin):
            self._log([u"Returning %.3f %.3f", begin, end])
            return (begin, end)
        self._log(u"Returning (0.0, 0.0)")
        return (0.0, 0.0)

    def detect_head(
            self,
            min_head_length=MIN_HEAD_LENGTH,
            max_head_length=MAX_HEAD_LENGTH,
            metric=SDMetric.VALUE
        ):
        """
        Detect the audio head.

        :param min_head_length: estimated minimum head length
        :type  min_head_length: float
        :param max_head_length: estimated maximum head length
        :type  max_head_length: float
        :param metric: the metric to be used when comparing candidates
        :type  metric: :class:`aeneas.sd.SDMetric`
        :rtype: float
        """
        return self._detect_head_or_tail(True, min_head_length, max_head_length, metric)

    def detect_tail(
            self,
            min_tail_length=MIN_TAIL_LENGTH,
            max_tail_length=MAX_TAIL_LENGTH,
            metric=SDMetric.VALUE
        ):
        """
        Detect the audio tail.

        :param min_tail_length: estimated minimum tail length
        :type  min_tail_length: float
        :param max_tail_length: estimated maximum tail length
        :type  max_tail_length: float
        :param metric: the metric to be used when comparing candidates
        :type  metric: :class:`aeneas.sd.SDMetric`
        :rtype: float
        """
        return self._detect_head_or_tail(False, min_tail_length, max_tail_length, metric)

    def _detect_head_or_tail(self, head, min_length, max_length, metric):
        """
        Detect the audio head or tail

        :param head: if ``True``, detect head; otherwise, detect tail
        :type  head: bool
        :param min_length: estimated minimum head/tail length
        :type  min_length: float
        :param max_length: estimated maximum head/tail length
        :type  max_length: float
        :param metric: the metric to be used when comparing candidates
        :type  metric: :class:`aeneas.sd.SDMetric`
        :rtype: float
        """
        if not head:
            self._log(u"Reversing audio first time")
            self.audio_file.reverse()
        self._extract_mfcc()
        self._extract_speech()
        if not head:
            self._log(u"Reversing audio second time")
            self.audio_file.reverse()
        self.audio_file.clear_data()
        value = 0.0
        try:
            value = self._detect_start(min_length, max_length, metric, not head)
        except Exception as exc:
            if head:
                self._log(u"An unexpected exception occurred while detecting head:", Logger.WARNING)
            else:
                self._log(u"An unexpected exception occurred while detecting tail:", Logger.WARNING)
            self._log([u"%s", exc], Logger.WARNING)
        return value

    # TODO simplify this function
    def _detect_start(self, min_start_length, max_start_length, metric, backwards=False):
        """ Detect start """

        self._log([u"Min start length: %.3f", min_start_length])
        self._log([u"Max start length: %.3f", max_start_length])
        self._log([u"Metric:           %s", metric])
        self._log([u"Backwards:        %s", str(backwards)])

        audio_rate = self.text_file.characters / self.audio_file.audio_length
        self._log([u"Audio rate:     %.3f", audio_rate])

        self._log(u"Synthesizing query...")
        tmp_handler, tmp_file_path = gf.tmp_file(suffix=u".wav", root=self.rconf["tmp_path"])
        synt = Synthesizer(rconf=self.rconf, logger=self.logger)
        synt_duration = max_start_length * self.QUERY_FACTOR
        self._log([u"Synthesizing %.3f seconds", synt_duration])
        result = synt.synthesize(
            self.text_file,
            tmp_file_path,
            quit_after=synt_duration,
            backwards=backwards
        )
        self._log(u"Synthesizing query... done")

        query_file = AudioFileMonoWAVE(tmp_file_path, rconf=self.rconf, logger=self.logger)
        if backwards:
            self._log(u"Reversing query")
            query_file.reverse()
        self._log(u"Extracting MFCCs for query...")
        query_file.extract_mfcc()
        query_file.clear_data()
        self._log(u"Extracting MFCCs for query... done")

        self._log(u"Cleaning up...")
        gf.delete_file(tmp_handler, tmp_file_path)
        self._log(u"Cleaning up... done")

        query_characters = result[2]
        query_len = query_file.audio_length
        query_mfcc = query_file.audio_mfcc
        query_rate = query_characters / query_len

        stretch_factor = max(1.0, query_rate / audio_rate)
        self._log([u"Audio rate:     %.3f", audio_rate])
        self._log([u"Query rate:     %.3f", query_rate])
        self._log([u"Stretch factor: %.3f", stretch_factor])

        audio_mfcc = self.audio_file.audio_mfcc
        self._log([u"Actual audio has %d frames", audio_mfcc.shape[1]])
        audio_mfcc_end_index = int(max_start_length * self.AUDIO_FACTOR / self.rconf["mfcc_win_shift"])
        self._log([u"Limiting audio to first %d frames", audio_mfcc_end_index])
        audio_mfcc_end_index = min(audio_mfcc_end_index, audio_mfcc.shape[1])
        audio_mfcc = audio_mfcc[:, 0:audio_mfcc_end_index]
        self._log([u"Limited audio has %d frames", audio_mfcc.shape[1]])

        l, o = audio_mfcc.shape
        l, n = query_mfcc.shape

        # minimum length of a matched interval in the real audio
        stretched_match_minimum_length = int(n * stretch_factor)

        self._log([u"Audio has %d frames == %.3f seconds", o, self._i2t(o)])
        self._log([u"Query has %d frames == %.3f seconds", n, self._i2t(n)])
        self._log([u"Stretch factor:          %.3f", stretch_factor])
        self._log([u"Required minimum length: %.3f", stretched_match_minimum_length])
        self._log(u"Speech intervals:")
        for interval in self.audio_speech:
            self._log([u"  %d %d == %.3f %.3f", self._t2i(interval[0]), self._t2i(interval[1]), interval[0], interval[1]])

        admissible_intervals = [x for x in self.audio_speech if ((x[0] >= min_start_length) and (x[0] <= max_start_length))]
        self._log(u"AdmissibleSpeech intervals:")
        for interval in admissible_intervals:
            self._log([u"  %d %d == %.3f %.3f", self._t2i(interval[0]), self._t2i(interval[1]), interval[0], interval[1]])

        candidates = []
        runs_with_min_length = 0
        runs_no_improvement = 0
        runs_min_distortion = numpy.inf
        runs_min_value = numpy.inf

        for interval in admissible_intervals:
            if runs_no_improvement >= self.MAX_RUNS_NO_IMPROVEMENT:
                self._log(u"  Breaking: too many runs without improvement")
                break

            if runs_with_min_length >= self.MAX_RUNS_WITH_MIN_LENGTH:
                self._log(u"  Breaking: too many runs with minimum required length")
                break

            start_time = interval[0]
            start_index = self._t2i(start_time)
            self._log([u"Evaluating interval starting at %d == %.3f ", start_index, start_time])
            if start_index > o:
                self._log(u"  Breaking: start index outside audio window")
                break

            req_end_index = start_index + stretched_match_minimum_length
            req_end_time = self._i2t(req_end_index)
            if req_end_index > o:
                self._log(u"  Breaking: not enough audio left in shifted window")
                break
            end_index = min(start_index + 2 * n, o)
            end_time = self._i2t(end_index)

            self._log([u"  Start   %d == %.3f", start_index, start_time])
            self._log([u"  Req end %d == %.3f", req_end_index, req_end_time])
            self._log([u"  Eff end %d == %.3f", end_index, end_time])

            audio_mfcc_sub = audio_mfcc[:, start_index:end_index]
            l, m = audio_mfcc_sub.shape

            self._log(u"Computing DTW...")
            aligner = DTWAligner(None, None, rconf=self.rconf, logger=self.logger)
            aligner.real_wave_full_mfcc = audio_mfcc_sub
            aligner.synt_wave_full_mfcc = query_mfcc
            aligner.real_wave_length = self._i2t(m)
            aligner.synt_wave_length = self._i2t(n)
            acm = aligner.compute_accumulated_cost_matrix()
            # transpose, so we have an n x m accumulated cost matrix
            acm = acm.transpose()
            last_row = acm[-1, :]
            self._log(u"Computing DTW... done")

            # find the minimum, but its index must be >= stretched_match_minimum_length
            candidate_argmin_index = numpy.argmin(last_row[stretched_match_minimum_length:])
            candidate_length_index = stretched_match_minimum_length + candidate_argmin_index
            candidate_length_time = self._i2t(candidate_length_index)
            candidate_value = last_row[candidate_length_index]
            candidate_end_index = start_index + candidate_length_index
            candidate_end_time = self._i2t(candidate_end_index)
            candidate_distortion = candidate_value / candidate_length_index

            # check if the candidate has minimum length
            if candidate_length_index == stretched_match_minimum_length:
                runs_with_min_length += 1
            else:
                runs_with_min_length = 0

            # check if the candidate improved the global minimum value
            if metric == SDMetric.VALUE:
                if candidate_value < runs_min_value:
                    runs_min_value = candidate_value
                    runs_no_improvement = 0
                else:
                    runs_no_improvement += 1
            if metric == SDMetric.DISTORTION:
                if candidate_distortion < runs_min_distortion:
                    runs_min_distortion = candidate_distortion
                    runs_no_improvement = 0
                else:
                    runs_no_improvement += 1

            # append to the list of candidates
            self._log([u"    Interval  start:      %d == %.6f", start_index, start_time])
            self._log([u"    Interval  end:        %d == %.6f", end_index, end_time])
            self._log([u"    Candidate start:      %d == %.6f", start_index, start_time])
            self._log([u"    Candidate end:        %d == %.6f", candidate_end_index, candidate_end_time])
            self._log([u"    Candidate length:     %d == %.6f", candidate_length_index, candidate_length_time])
            self._log([u"    Candidate value:      %.6f", candidate_value])
            self._log([u"    Candidate distortion: %.6f", candidate_distortion])
            candidates.append({
                "start_index": start_index,
                "length": candidate_length_index,
                "value": candidate_value,
                "distortion": candidate_distortion
            })

        # select best candidate and return its start time
        # if we have no best candidate, return 0.0
        best_candidate = self._select_best_candidate(candidates, metric)
        if best_candidate is None:
            return 0.0
        sd_time = self._i2t(max(best_candidate["start_index"], 0))
        self._log([u"Returning time %.3f", sd_time])
        return sd_time

    def _extract_mfcc(self):
        """ Extract MFCCs for audio """
        self._log(u"Extracting MFCCs for audio...")
        self.audio_file.extract_mfcc()
        self._log(u"Extracting MFCCs for audio... done")

    def _extract_speech(self):
        """ Extract speech intervals """
        self._log(u"Running VAD...")
        vad = VAD(
            self.audio_file.audio_mfcc,
            self.audio_file.audio_length,
            rconf=self.rconf,
            logger=self.logger
        )
        vad.compute_vad()
        self.audio_speech = vad.speech
        self._log(u"Running VAD... done")

    def _i2t(self, index):
        """ Frame index to (start) time """
        return int(index * self.rconf["mfcc_win_shift"])

    def _t2i(self, time):
        """ Frame (start) time to index """
        return int(time / self.rconf["mfcc_win_shift"])

    def _select_best_candidate(self, candidates, metric):
        """ Select the best candidate (or None if no one is found) """
        self._log([u"Using metric '%s'", metric])
        self._log(u"Candidates:")
        for candidate in candidates:
            self._log([u"  %s", str(candidate)])
        tuples = []
        if metric == SDMetric.VALUE:
            tuples = [(v["value"], v["distortion"], v) for v in candidates]
        if metric == SDMetric.DISTORTION:
            tuples = [(v["distortion"], v["value"], v) for v in candidates]
        if len(tuples) == 0:
            return None
        return min(tuples)[2]






#!/usr/bin/env python
# coding=utf-8

"""
The runtime configuration object.
"""

from __future__ import absolute_import
from __future__ import print_function

from aeneas.configurationobject import ConfigurationObject
import aeneas.globalconstants as gc

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class RuntimeConfiguration(ConfigurationObject):
    """
    A structure representing a runtime configuration, that is,
    a set of parameters for the algorithms which process Jobs and Tasks.

    Allowed keys:

    * ``gc.XYZ`` or ``.xyz``

    :param config_string: the configuration string
    :type  config_string: Unicode string

    :raises TypeError: if ``config_string`` is not ``None`` and
                       it is not a Unicode string
    :raises KeyError: if trying to access a key not listed above
    """

    TAG = u"RuntimeConfiguration"

    FIELDS = [
        (gc.RC_ALLOW_UNLISTED_LANGUAGES, (False, bool, ["allow_unlisted_languages"])),

        (gc.RC_C_EXTENSIONS, (True, bool, ["c_ext"])),

        (gc.RC_DTW_ALGORITM, ("stripe", None, ["dtw_algorithm"])),
        (gc.RC_DTW_MARGIN, (60.0, float, ["dtw_margin"])),

        #(gc.RC_ESPEAK_PATH, ("/usr/bin/espeak", None, ["espeak_path"])),
        (gc.RC_ESPEAK_PATH, ("espeak", None, ["espeak_path"])),

        #(gc.RC_FFMPEG_PATH, ("/usr/bin/ffmpeg", None, ["ffmpeg_path"])),
        (gc.RC_FFMPEG_PATH, ("ffmpeg", None, ["ffmpeg_path"])),
        (gc.RC_FFMPEG_SAMPLE_RATE, (16000, int, ["ffmpeg_sr"])),

        #(gc.RC_FFMPEG_PATH, ("/usr/bin/ffprobe", None, ["ffprobe_path"])),
        (gc.RC_FFPROBE_PATH, ("ffprobe", None, ["ffprobe_path"])),

        (gc.RC_JOB_MAX_TASKS, (0, int, ["job_max_tasks"])),

        (gc.RC_MFCC_FILTERS, (40, int, ["mfcc_filters"])),
        (gc.RC_MFCC_SIZE, (13, int, ["mfcc_size"])),
        (gc.RC_MFCC_FFT_ORDER, (512, int, ["mfcc_order"])),
        (gc.RC_MFCC_LOWER_FREQUENCY, (133.3333, float, ["mfcc_lower_freq"])),
        (gc.RC_MFCC_UPPER_FREQUENCY, (6855.4976, float, ["mfcc_upper_freq"])),
        (gc.RC_MFCC_EMPHASIS_FACTOR, (0.970, float, ["mfcc_emph"])),
        (gc.RC_MFCC_WINDOW_LENGTH, (0.100, float, ["mfcc_win_len"])),
        (gc.RC_MFCC_WINDOW_SHIFT, (0.040, float, ["mfcc_win_shift"])),

        (gc.RC_TASK_MAX_AUDIO_LENGTH, (7200.0, float, ["task_max_a_len"])),
        (gc.RC_TASK_MAX_TEXT_LENGTH, (0, int, ["task_max_t_len"])),

        (gc.RC_TMP_PATH, (None, None, ["tmp_path"])),

        (gc.RC_VAD_EXTEND_SPEECH_INTERVAL_AFTER, (0.0, float, ["vad_extend_s_after"])),
        (gc.RC_VAD_EXTEND_SPEECH_INTERVAL_BEFORE, (0.0, float, ["vad_extend_s_before"])),
        (gc.RC_VAD_LOG_ENERGY_THRESHOLD, (0.699, float, ["vad_log_energy_thr"])),
        (gc.RC_VAD_MIN_NONSPEECH_LENGTH, (0.200, float, ["vad_min_ns_len"])),
    ]

    def __init__(self, config_string=None):
        super(RuntimeConfiguration, self).__init__(config_string)






#!/usr/bin/env python
# coding=utf-8

"""
Download a file from a Web source.

Currently, it downloads an audio file from a YouTube video.
"""

from __future__ import absolute_import
from __future__ import print_function
import sys

from aeneas.downloader import Downloader
from aeneas.tools.abstract_cli_program import AbstractCLIProgram
import aeneas.globalfunctions as gf

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL 3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class DownloadCLI(AbstractCLIProgram):
    """
    Download a file from a Web source.

    Currently, it downloads an audio file from a YouTube video.
    """
    OUTPUT_FILE_M4A = "output/sonnet.m4a"
    OUTPUT_FILE_OGG = "output/sonnet.ogg"
    URL_YOUTUBE = "https://www.youtube.com/watch?v=rU4a7AA8wM0"

    NAME = gf.file_name_without_extension(__file__)

    HELP = {
        "description": u"Download an audio file from a YouTube video.",
        "synopsis": [
            u"YOUTUBE_URL [OUTPUT_FILE]"
        ],
        "examples": [
            u"%s --list" % (URL_YOUTUBE),
            u"%s %s" % (URL_YOUTUBE, OUTPUT_FILE_M4A),
            u"%s %s --index=0" % (URL_YOUTUBE, OUTPUT_FILE_M4A),
            u"%s %s --smallest-audio" % (URL_YOUTUBE, OUTPUT_FILE_OGG),
            u"%s %s --largest-audio --format=ogg" % (URL_YOUTUBE, OUTPUT_FILE_OGG),
        ],
        "options": [
            u"--format=FMT : preferably download audio stream in FMT format",
            u"--index=IDX : download audio stream with given index",
            u"--largest-audio : download largest audio stream (default)",
            u"--list : list all available audio streams but do not download",
            u"--smallest-audio : download smallest audio stream",
        ]
    }

    def perform_command(self):
        """
        Perform command and return the appropriate exit code.

        :rtype: int
        """
        if len(self.actual_arguments) < 2:
            return self.print_help()
        source_url = self.actual_arguments[0]
        output_file_path = self.actual_arguments[1]

        download = not self.has_option("--list")
        # largest_audio = True by default or if explicitly given
        if self.has_option("--largest-audio"):
            largest_audio = True
        else:
            largest_audio = not self.has_option("--smallest-audio")
        preferred_format = self.has_option_with_value("--format")
        preferred_index = gf.safe_int(self.has_option_with_value("--index"), None)

        try:
            if download:
                self.print_info(u"Downloading audio stream from '%s' ..." % source_url)
                downloader = Downloader(logger=self.logger)
                result = downloader.audio_from_youtube(
                    source_url,
                    download=download,
                    output_file_path=output_file_path,
                    preferred_index=preferred_index,
                    largest_audio=largest_audio,
                    preferred_format=preferred_format
                )
                self.print_info(u"Downloading audio stream from '%s' ... done" % source_url)
                self.print_info(u"Downloaded file '%s'" % result)
            else:
                self.print_info(u"Downloading stream info from '%s' ..." % source_url)
                downloader = Downloader(logger=self.logger)
                result = downloader.audio_from_youtube(
                    source_url,
                    download=False
                )
                self.print_info(u"Downloading stream info from '%s' ... done" % source_url)
                msg = []
                msg.append(u"%s\t%s\t%s\t%s" % ("Index", "Format", "Bitrate", "Size"))
                i = 0
                for audio in result:
                    ext = audio.extension
                    bitrate = audio.bitrate
                    size = gf.human_readable_number(audio.get_filesize())
                    msg.append(u"%d\t%s\t%s\t%s" % (i, ext, bitrate, size))
                    i += 1
                self.print_generic(u"Available audio streams:")
                self.print_generic(u"\n".join(msg))
            return self.NO_ERROR_EXIT_CODE
        except ImportError:
            self.print_error(u"You need to install Python module pafy to download audio from YouTube. Run:")
            self.print_error(u"$ sudo pip install pafy")
        except Exception as exc:
            self.print_error(u"An unexpected Exception occurred while downloading audio from YouTube:")
            self.print_error(u"%s" % exc)

        return self.ERROR_EXIT_CODE


def main():
    """
    Execute program.
    """
    DownloadCLI().run(arguments=sys.argv)

if __name__ == '__main__':
    main()






#!/usr/bin/env python
# coding=utf-8

"""
Enumeration of the available algorithms to sort IDs.
"""

from __future__ import absolute_import
from __future__ import print_function
import re

from aeneas.logger import Logger
from aeneas.runtimeconfiguration import RuntimeConfiguration

__author__ = "Alberto Pettarin"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2016, Alberto Pettarin (www.albertopettarin.it)
    """
__license__ = "GNU AGPL v3"
__version__ = "1.4.1"
__email__ = "aeneas@readbeyond.it"
__status__ = "Production"

class IDSortingAlgorithm(object):
    """
    Enumeration of the available algorithms to sort
    a list of XML ``id`` attributes.

    :param algorithm: the id sorting algorithm to be used
    :type  algorithm: :class:`aeneas.idsortingalgorithm.IDSortingAlgorithm`
    :param rconf: a runtime configuration. Default: ``None``, meaning that
                  default settings will be used.
    :type  rconf: :class:`aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`aeneas.logger.Logger`

    :raise ValueError: if the value of ``algorithm`` is not allowed
    """

    LEXICOGRAPHIC = "lexicographic"
    """ Lexicographic sorting
    (e.g., ``f020`` before ``f10`` before ``f2``) """

    NUMERIC = "numeric"
    """ Numeric sorting, ignoring any non-digit characters or leading zeroes
    (e.g., ``f2`` (= ``2``) before ``f10`` (= ``10``) before ``f020`` (= ``20``)) """

    UNSORTED = "unsorted"
    """ Do not sort and return the list of identifiers, in their original order
    (e.g., ``f2`` before ``f020`` before ``f10``,
    assuming this was their order in the XML DOM) """

    ALLOWED_VALUES = [LEXICOGRAPHIC, NUMERIC, UNSORTED]
    """ List of all the allowed values """

    TAG = u"IDSortingAlgorithm"

    def __init__(self, algorithm, rconf=None, logger=None):
        if algorithm not in self.ALLOWED_VALUES:
            raise ValueError("Algorithm value not allowed")
        self.algorithm = algorithm
        self.logger = logger or Logger()
        self.rconf = rconf or RuntimeConfiguration()

    def _log(self, message, severity=Logger.DEBUG):
        """ Log """
        self.logger.log(message, severity, self.TAG)

    def sort(self, ids):
        """
        Sort the given list of identifiers,
        returning a new (sorted) list.

        :param ids: the list of identifiers to be sorted
        :type ids: list of Unicode strings
        :rtype: list of Unicode strings
        """
        def extract_int(string):
            """
            Extract an integer from the given string.

            :param string: the identifier string
            :type  string: string
            :rtype: int
            """
            return int(re.sub(r"[^0-9]", "", string))

        tmp = list(ids)
        if self.algorithm == IDSortingAlgorithm.UNSORTED:
            self._log(u"Sorting using UNSORTED")
        elif self.algorithm == IDSortingAlgorithm.LEXICOGRAPHIC:
            self._log(u"Sorting using LEXICOGRAPHIC")
            tmp = sorted(ids)
        elif self.algorithm == IDSortingAlgorithm.NUMERIC:
            self._log(u"Sorting using NUMERIC")
            tmp = ids
            try:
                tmp = sorted(tmp, key=extract_int)
            except (ValueError, TypeError) as exc:
                self._log(u"Not all id values contain a numeric part:", Logger.WARNING)
                self._log([u"%s", exc], Logger.WARNING)
                self._log(u"Returning the id list unchanged (unsorted)", Logger.WARNING)
        return tmp



